define(['underscore'], function (_) {

    //继承相关逻辑
    (function () {

        // 全局可能用到的变量
        var arr = [];
        var slice = arr.slice;
        /**
         * inherit方法，js的继承，默认为两个参数
         *
         * @param  {function} origin  可选，要继承的类
         * @param  {object}   methods 被创建类的成员，扩展的方法和属性
         * @return {function}         继承之后的子类
         */
        _.inherit = function (origin, methods) {

            // 参数检测，该继承方法，只支持一个参数创建类，或者两个参数继承类
            if (arguments.length === 0 || arguments.length > 2) throw '参数错误';

            var parent = null;

            // 将参数转换为数组
            var properties = slice.call(arguments);

            // 如果第一个参数为类（function），那么就将之取出
            if (typeof properties[0] === 'function')
                parent = properties.shift();
            properties = properties[0];

            // 创建新类用于返回
            function klass() {
                if (_.isFunction(this.initialize))
                    this.initialize.apply(this, arguments);
            }

            klass.superclass = parent;

            // 父类的方法不做保留，直接赋给子类
            // parent.subclasses = [];

            if (parent) {
                // 中间过渡类，防止parent的构造函数被执行
                var subclass = function () {
                };
                subclass.prototype = parent.prototype;
                klass.prototype = new subclass();

                // 父类的方法不做保留，直接赋给子类
                // parent.subclasses.push(klass);
            }

            var ancestor = klass.superclass && klass.superclass.prototype;
            for (var k in properties) {
                var value = properties[k];

                //满足条件就重写
                if (ancestor && typeof value == 'function') {
                    var argslist = /^\s*function\s*\(([^\(\)]*?)\)\s*?\{/i.exec(value.toString())[1].replace(/\s/g, '').split(',');
                    //只有在第一个参数为$super情况下才需要处理（是否具有重复方法需要用户自己决定）
                    if (argslist[0] === '$super' && ancestor[k]) {
                        value = (function (methodName, fn) {
                            return function () {
                                var scope = this;
                                var args = [
                                    function () {
                                        return ancestor[methodName].apply(scope, arguments);
                                    }
                                ];
                                return fn.apply(this, args.concat(slice.call(arguments)));
                            };
                        })(k, value);
                    }
                }

                //此处对对象进行扩展，当前原型链已经存在该对象，便进行扩展
                if (_.isObject(klass.prototype[k]) && _.isObject(value) && (typeof klass.prototype[k] != 'function' && typeof value != 'fuction')) {
                    //原型链是共享的，这里处理逻辑要改
                    var temp = {};
                    _.extend(temp, klass.prototype[k]);
                    _.extend(temp, value);
                    klass.prototype[k] = temp;
                } else {
                    klass.prototype[k] = value;
                }
            }

            //静态属性继承
            //兼容代码，非原型属性也需要进行继承
            for (key in parent) {
                if (parent.hasOwnProperty(key) && key !== 'prototype' && key !== 'superclass')
                    klass[key] = parent[key];
            }

            if (!klass.prototype.initialize)
                klass.prototype.initialize = function () {
                };

            klass.prototype.constructor = klass;

            return klass;
        };

    })();

    var Storage = _.inherit({
        //默认属性
        propertys: function () {

            //代理对象，默认为localstorage
            this.sProxy = window.localStorage;

            //解决隐私模式不可用问题，会很大程度影响执行效率
            try {
                window.localStorage.setItem('test', 'test');
            } catch (e) {
                this.sProxy.setItem = function () {
                    return;
                };
                this.sProxy.getItem = function () {
                    return null;
                };
            }

            //60 * 60 * 24 * 30 * 1000 ms ==30天
            this.defaultLifeTime = 2592000000;

            //本地缓存用以存放所有localstorage键值与过期日期的映射
            this.keyCache = 'SYSTEM_KEY_TIMEOUT_MAP';

            //当缓存容量已满，每次删除的缓存数
            this.removeNum = 5;

        },

        assert: function () {
            if (this.sProxy === null) {
                throw 'not override sProxy property';
            }
        },

        initialize: function (opts) {
            this.propertys();
            this.assert();
        },

        /*
        新增localstorage
        数据格式包括唯一键值，json字符串，过期日期，存入日期
        sign 为格式化后的请求参数，用于同一请求不同参数时候返回新数据，比如列表为北京的城市，后切换为上海，会判断tag不同而更新缓存数据，tag相当于签名
        每一键值只会缓存一条信息
        */
        set: function (key, value, timeout, sign, shoudClear) {
            var _d = new Date();
            //存入日期
            var indate = _d.getTime();

            //最终保存的数据
            var entity = null;

            if (!timeout) {
                _d.setTime(_d.getTime() + this.defaultLifeTime);
                timeout = _d.getTime();
            }

            //
            this.setKeyCache(key, timeout);
            entity = this.buildStorageObj(value, indate, timeout, sign, shoudClear);

            try {
                this.sProxy.setItem(key, JSON.stringify(entity));
                return true;
            } catch (e) {
                //localstorage写满时,全清掉
                if (e.name == 'QuotaExceededError') {
                    //            this.sProxy.clear();
                    //localstorage写满时，选择离过期时间最近的数据删除，这样也会有些影响，但是感觉比全清除好些，如果缓存过多，此过程比较耗时，100ms以内
                    if (!this.removeLastCache()) throw '本次数据存储量过大';
                    this.set(key, value, timeout, sign);
                }
                console && console.log(e);
            }
            return false;
        },

        //删除过期缓存
        removeOverdueCache: function () {
            var tmpObj = null, i, len;

            var now = new Date().getTime();
            //取出键值对
            var cacheStr = this.sProxy.getItem(this.keyCache);
            var cacheMap = [];
            var newMap = [];
            if (!cacheStr) {
                return;
            }

            cacheMap = JSON.parse(cacheStr);

            for (i = 0, len = cacheMap.length; i < len; i++) {
                tmpObj = cacheMap[i];
                if (tmpObj.timeout < now) {
                    this.sProxy.removeItem(tmpObj.key);
                } else {
                    newMap.push(tmpObj);
                }
            }
            this.sProxy.setItem(this.keyCache, JSON.stringify(newMap));

        },

        removeLastCache: function () {
            var i, len;
            var num = this.removeNum || 5;

            //取出键值对
            var cacheStr = this.sProxy.getItem(this.keyCache);
            var cacheMap = [];
            var delMap = [];

            //说明本次存储过大
            if (!cacheStr) return false;

            cacheMap.sort(function (a, b) {
                return a.timeout - b.timeout;
            });

            //删除了哪些数据
            delMap = cacheMap.splice(0, num);
            for (i = 0, len = delMap.length; i < len; i++) {
                this.sProxy.removeItem(delMap[i].key);
            }

            this.sProxy.setItem(this.keyCache, JSON.stringify(cacheMap));
            return true;
        },

        setKeyCache: function (key, timeout) {
            if (!key || !timeout || timeout < new Date().getTime()) return;
            var i, len, tmpObj;

            //获取当前已经缓存的键值字符串
            var oldstr = this.sProxy.getItem(this.keyCache);
            var oldMap = [];
            //当前key是否已经存在
            var flag = false;
            var obj = {};
            obj.key = key;
            obj.timeout = timeout;

            if (oldstr) {
                oldMap = JSON.parse(oldstr);
                if (!_.isArray(oldMap)) oldMap = [];
            }

            for (i = 0, len = oldMap.length; i < len; i++) {
                tmpObj = oldMap[i];
                if (tmpObj.key == key) {
                    oldMap[i] = obj;
                    flag = true;
                    break;
                }
            }
            if (!flag) oldMap.push(obj);
            //最后将新数组放到缓存中
            this.sProxy.setItem(this.keyCache, JSON.stringify(oldMap));

        },

        buildStorageObj: function (value, indate, timeout, sign, shouldClear) {
            var obj = {
                value: value,
                timeout: timeout,
                sign: sign,
                indate: indate
            };

            if(shouldClear) obj.shouldClear = true;
            return obj;
        },

        get: function (key, sign) {
            var result, now = new Date().getTime();
            try {
                result = this.sProxy.getItem(key);
                if (!result) return null;
                result = JSON.parse(result);

                //数据过期
                if (result.timeout < now) return null;

                //需要验证签名
                if (sign) {
                    if (sign === result.sign)
                        return result.value;
                    return null;
                } else {
                    return result.value;
                }

            } catch (e) {
                console && console.log(e);
            }
            return null;
        },

        //获取签名
        getSign: function (key) {
            var result, sign = null;
            try {
                result = this.sProxy.getItem(key);
                if (result) {
                    result = JSON.parse(result);
                    sign = result && result.sign
                }
            } catch (e) {
                console && console.log(e);
            }
            return sign;
        },

        remove: function (key) {
            return this.sProxy.removeItem(key);
        },

        clear: function () {
            this.sProxy.clear();
        }
    });

    Storage.clearData = function () {
        var data = {};
        var removeKeys = [];
        var str = '';
        for(var k in localStorage) {
            if(k == 'SYSTEM_KEY_TIMEOUT_MAP' || k == 'test') continue;
            str = localStorage.getItem(k);

            if(str.charAt(0) != '[' && str.charAt(0) != '{') continue;

            data = JSON.parse(str);
            if(data.shouldClear) {
                removeKeys.push(k);
            }
        }

        for(var i = 0; i < removeKeys.length; i++) {
            localStorage.removeItem(removeKeys[i]);
        }

    };

    Storage.getInstance = function () {
        if (this.instance) {
            return this.instance;
        } else {
            return this.instance = new this();
        }
    };

    return Storage;

});
